# AdventOfCode2024

Solver for Advent of Code 2024 puzzles. <https://adventofcode.com/2024/>

This is a single command line application for the puzzles written in [C++23](https://en.cppreference.com/w/cpp/23) with CMake.

## Puzzle Input

This project does not contain the puzzle or example inputs as per the [copyright notice of Advent of Code](https://adventofcode.com/about). In order to run the compiled application, the puzzle inputs have to be downloaded from the [Advent of Code 2024](https://adventofcode.com/2024/) puzzle pages, and placed as text files into the `AdventOfCode2024\data` directory, e.g. `AdventOfCode2024\data\historian_hysteria.txt`, or `AdventOfCode2024\data\example\historian_hysteria.txt` for the unit tests. The application will output an error message with details if it cannot find an input file.

## Tests

The solution contains a unit test project to help troubleshoot issues and prevent regressions in the solver class framework. These tests cover the solutions for provided examples and full data inputs. I did not find a definitive source for it, but the full data inputs seem to be user-specific, so adding my solutions as tests does not spoil the solutions.

## Solutions

### Day 1: Historian Hysteria

:mag_right: Puzzle: <https://adventofcode.com/2024/day/1>, :white_check_mark: Solver: [`HistorianHysteria.cpp`](src/HistorianHysteria.cpp)

I'm using a `std::multiset` to collect and sort the values for both lists. This allows to use a single iteration of the left list and two iterations of the right list simultaneously to solve both parts. Nice application of iterators.

### Day 2: Red-Nosed Reports

:mag_right: Puzzle: <https://adventofcode.com/2024/day/2>, :white_check_mark: Solver: [`RedNosedReports.cpp`](src/RedNosedReports.cpp)

Here, we have a few conditionals to determine on the fly which of the numbers would make the report safe if dropped. The amount of cases is actually quite manageable.

### Day 3: Mull It Over

:mag_right: Puzzle: <https://adventofcode.com/2024/day/3>, :white_check_mark: Solver: [`MullItOver.cpp`](src/MullItOver.cpp)

A simple [finite state machine](include/aoc/common/StringStateMachine.hpp) crawling along the input character by character solves both parts nicely. The algorithm tracks whether `mul` instructions are enabled or not, but ignores this setting for part 1.

### Day 4: Ceres Search

:mag_right: Puzzle: <https://adventofcode.com/2024/day/4>, :white_check_mark: Solver: [`CeresSearch.cpp`](src/CeresSearch.cpp)

For this puzzle I added a class for [points in two-dimensional space](include/aoc/common/Point2.hpp), so I can use these for simplifying directional computations. With that, the algorithm looks for all `X` and `A` for part 1 and 2, respectively, and tries to find the remaining characters, starting from that `X` or `A`.

### Day 5: Print Queue

:mag_right: Puzzle: <https://adventofcode.com/2024/day/5>, :white_check_mark: Solver: [`PrintQueue.cpp`](src/PrintQueue.cpp)

My implementation uses an ordering matrix (a two-dimensional boolean array) to track which page combinations are ordered, and then queries that matrix for each ordered combination of pages in a single line. The same matrix can then also be used for a custom sort function for part 2.

### Day 6: Guard Gallivant

:mag_right: Puzzle: <https://adventofcode.com/2024/day/6>, :white_check_mark: Solver: [`GuardGallivant.cpp`](src/GuardGallivant.cpp)

The solver for this puzzle simply simulates the guard traversing the map, turning them right at obstacles and counting the unique positions the guard will visit for part one. At each newly visited position, the algorithm will place a temporary obstacle at the next empty position and check if that obstacle will create a cycle for part two.

In order to improve performance for part two, the solver backtraces at turns to find all steps on the map that would lead to the position the guard is currently in, and uses this data to find cycles early.

### Day 7: Bridge Repair

:mag_right: Puzzle: <https://adventofcode.com/2024/day/7>, :white_check_mark: Solver: [`BridgeRepair.cpp`](src/BridgeRepair.cpp)

The algorithm recursively tries the different operators from left to right until all the calibration numbers have been used. Since the result from any of the operator cannot be less than its operands, any branch of the recursion that surpasses the test value can be aborted.

### Day 8: Resonant Collinearity

:mag_right: Puzzle: <https://adventofcode.com/2024/day/8>, :white_check_mark: Solver: [`ResonantCollinearity.cpp`](src/ResonantCollinearity.cpp)

For any pair of antennas of the same frequency, the algorithm adds the difference of their positions to one of these positions and subtracts it from the other to find the antinodes. For part 2, this process is repeated for each of the two directions until the found antinodes leave the defined area. Counting duplicate antinodes is prevented by keeping track of their locations in a two-dimensional array covering the area.

### Day 9: Disk Fragmenter

:mag_right: Puzzle: <https://adventofcode.com/2024/day/9>, :white_check_mark: Solver: [`DiskFragmenter.cpp`](src/DiskFragmenter.cpp)

The input string is processed simultaneously from both ends to calculate the checksums of files or file blocks without actually moving them. The algorithm does not have to keep track of positions for specific files or file blocks. For part two, a list of empty spaces is maintained, but only up to the point where the front and back processing meets. An array of indices per possible file size, where each index tracks the first empty space that could hold a file of that size and advances as needed, is also kept to speed up the search in the list of empty spaces.

### Day 10: Hoof It

:mag_right: Puzzle: <https://adventofcode.com/2024/day/10>, :white_check_mark: Solver: [`HoofIt.cpp`](src/HoofIt.cpp)

The algorithm starts at the trail ends (the `9`'s) and searches all options from there downwards until trail heads are reached. For each visited map location, it tracks unique reachable end points (for part 1) and number of possible paths to end points (for part 2). These values are accumulated, if during the search the same map location is encountered again.

### Day 11: Plutonian Pebbles

:mag_right: Puzzle: <https://adventofcode.com/2024/day/11>, :white_check_mark: Solver: [`PlutonianPebbles.cpp`](src/PlutonianPebbles.cpp)

Since the order of the pebbles does not actually matter, we can simply blink from one line of pebbles to the next and count only cardinalities of each number in the current line.

### Day 12: Garden Groups

:mag_right: Puzzle: <https://adventofcode.com/2024/day/12>, :white_check_mark: Solver: [`GardenGroups.cpp`](src/GardenGroups.cpp)

The algorithm uses two stacks to flood-fill each garden region; one stack for all plots inside the region, and one stack to collect all other plots encountered during the flood-fill. Once a region has been traversed, it just pops the next plot from the latter stack. This gives us naturally the area of a region.

The perimeter of a region is the sum of neighbors for all plots encountered during the traversal, that are not in the same region, also counting neighbors outside the defined area. The number of sides of a region is equal to the number of corners of a region, which can also be counted per plot while doing the region traversal. A corner occurs for each pair of adjacent cardinal neighbors if they are either both outside of the current region, or if they are both inside the region and the diagonal neighbor in between them is not.

Adding up the products of area and perimeter, or area and corner count, per region gives us the solution.

### Day 13: Claw Contraption

:mag_right: Puzzle: <https://adventofcode.com/2024/day/13>, :white_check_mark: Solver: [`ClawContraption.cpp`](src/ClawContraption.cpp)

For each claw machine, we have to solve a simple linear equation system to get the number of required button pushes. Conveniently, for the given data, all of these systems have a unique solution. However, some solutions have to be excluded because the number of button pushes is not integer, or because the solutions are too high for part 1. The required integer tests are done by performing integer division, and comparing the dividend with the product of the integer quotient and the divisor.

### Day 14: Restroom Redoubt

:mag_right: Puzzle: <https://adventofcode.com/2024/day/14>, :white_check_mark: Solver: [`RestroomRedoubt.cpp`](src/RestroomRedoubt.cpp)

For part 1, the resulting quadrant for each robot can be calculated directly, by multiplying the velocity vector by 100 and adding it to the position vector, modulo width or height to stay within the target range.

When observing a few resulting robot configurations after each passed second, it becomes apparent that some of them show vertical or horizontal bands of clustering, and that these repeat every `width` or `height` seconds, respectively. With a bit of modulo calculus, the time after which the vertical and horizontal bands overlap can be determined, which is the solution for part 2.

### Day 15: Warehouse Woes

:mag_right: Puzzle: <https://adventofcode.com/2024/day/15>, :white_check_mark: Solver: [`WarehouseWoes.cpp`](src/WarehouseWoes.cpp)

The idea for this solver was pretty straight forward. The algorithm moves the robot around the warehouse following the instructions, and checks whether boxes in its way can be pushed, that is whether there is an empty spot behind the box, or stack of boxes.

It is more complicated for part 2 of course, since here horizontal and vertical directions behave differently. For vertical pushes there are two phases, the first one where individual boxes to be pushed are found, tracked, and checked for walls behind them, and a second phase where all those boxes are moved in reverse order.

## Thanks

* [Alexander Brouwer](https://github.com/Bromvlieg) for getting the project set up with CMake.

## License

Copyright (C) 2024 Stefan Müller

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
